#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#include "linked_list.h"

// A linked list, defined as a global variable.
linked_list_t   list;

// A pthread mutex that should be used to protect access to the list.
pthread_mutex_t list_mutex;


// This function executes a computationally bound spinloop, and also
// injects something into the linked list "list".
void *spinloop(void *arg) {
  int argument = (int) arg;
  unsigned long long int i, j = 0;

  printf("%d: starting spinloop\n", argument);
  for (i=0ULL; i<5000000000ULL; i++) {
    j+=i;
  }
  printf("%d: finished spinloop\n", argument);

  // Need to wrap access to the linked list with a lock, otherwise
  // multiple threads might enter linked_list_add_to_back() simultaneously
  // and try to operate on the internal data structures of the list,
  // causing all sorts of havoc.
  if (pthread_mutex_lock(&list_mutex) != 0) {
    printf("pthread_mutex_lock() failed! exiting process\n");
    exit(-1);
  }
  if (!linked_list_add_to_back(list, (list_element_t) argument)) {
    printf("Failed to add to linked list; exiting process\n");
    exit(-1);
  }
  if (pthread_mutex_unlock(&list_mutex) != 0) {
    printf("pthread_mutex_unlock() failed! exiting process\n");
    exit(-1);
  }

  return NULL;
}

// The process begins executing here.  When the process starts
// executing, it has a single thread of execution.
int main(int argc, char **argv) {
  pthread_t thread1, thread2;

  // Allocated / initialize the list.
  list = linked_list_create();
  if (list == NULL) {
    printf("Failed to create the linked list\n");
    return -1;
  }

  // Initialize the pthread_mutex.
  if (pthread_mutex_init(&list_mutex, NULL) != 0) {
    printf("pthread_mutex_init() failed\n");
    return -1;
  }

  // Create two threads, and dispatch them into spinloop.
  if (pthread_create(&thread1, NULL, spinloop, (void *) 1) != 0) {
    printf("pthread_create() of thread id 1 failed; exiting process\n");
    return -1;
  }
  if (pthread_create(&thread2, NULL, spinloop, (void *) 2) != 0) {
    printf("pthread_create() of thread id 2 failed; exiting process\n");
    return -1;
  }

  // Three threads now exist (main, thread1, thread2), unless either of
  // thread1/thread2 have already finished executing.  thread1 and thread2
  // are executing in the "spinloop" function, and the main thread
  // is executing here.

  // Wait for those two threads to terminate.  Note that the pthreads
  // library is smart enough to handle a NULL pointer as a second argument;
  // normally, pthreads will shove the return value from the thread's
  // dispatch function into whatever the second argument points to, but
  // if the second argument is NULL, it won't try to do this.
  if (pthread_join(thread1, NULL) != 0) {
    printf("pthread_join() of thread id 1 failed; exiting process\n");
    return -1;
  }
  if (pthread_join(thread2, NULL) != 0) {
    printf("pthread_join() of thread id 2 failed; exiting process\n");
    return -1;
  }
  return 0;  // Returning 0 indicates successful execution.
}
